#!/bin/bash

# Personalize depending on the name of the modules 
# to be used in the compilation of the serial and 
# parallel version of DIABLO
# ---------------------------------------------------
MODULE_SERIAL="matlab hdf5/ifort  fftw2/intel/64/double/2.1.5"
MODULE_PARALLEL="matlab hdf5/impi-64b fftw2/intel/64/double/2.1.5"
MYPYTHONPATH="/Users/deusebio/.ipython/profile_default/startup/"
# ---------------------------------------------------

. ~/.bashrc

echo " "
echo "  ****************************************** "
echo "          SUITE FOR TESTING DIABLO "
echo "  ****************************************** "
echo " "

GetNumberNy(){
    local FileIn=$1
    local NY=$(sed -n 5p $FileIn)
    NY=${NY%%)*}
    NY=${NY##*=}
    echo $NY
}

HHOME=$(pwd)
export PATH_DIABLO=$HHOME"/../"
export PYTHONPATH=$PYTHONPATH":"$PATH_DIABLO"/post_process/python"
export PYTHONPATH=$PYTHONPATH":"$MYPYTHONPATH

module add $MODULE_SERIAL

# ---------------------
# HDF5 Folder
# ---------------------
if [ -d hdf5 ]; then rm -r hdf5; fi
mkdir hdf5;cd hdf5

# ---------------------
# STABILITY ANALYSIS
# ---------------------
if [ -d stab_chan ]; then rm -r stab_chan; fi
mkdir stab_chan;cd stab_chan

cp $HHOME"/data/grid_def" .
# ---------------------
# Generate HDF5 grid
# ---------------------
printf " Creating HDF5 grid ..."
cp $HHOME"/../pre_process/create_grid_h5.m" $HHOME"/../pre_process/matlab_basher.sh" .
if [ -f grid.h5 ]; then rm grid.h5; fi
NY=$(GetNumberNy grid_def)
./matlab_basher.sh create_grid_h5.m "GRID_TYPE=1;N="$NY";L=2;CS=1.75;" 
rm create_grid_h5.m 
printf "OK\n"
# ---------------------

# ---------------------
# Check for proper compilation 
# ---------------------
printf " Check for proper compilation ..."
cp $HHOME"/../bin/compDia.sh" . 
if [ -f diablo ]; then rm diablo; fi
./compDia.sh -SimDir=$HHOME"/../for" -Clean=yes -Options="HDF5=TRUE" > log.out
if [ ! -f diablo ]; then echo " Error 4343445. - Cannot compile serial version of diablo"; exit 1; fi
printf "OK\n"

# ---------------------
# Setup linear stability analysis simulation
# ---------------------
printf " Check for linear stability analysis ... \n"
cp $HHOME"/data/input.dat.stab" input.dat 
cp $HHOME"/data/input_chan.dat.stab" input_chan.dat 
# Create flow fields with matlab basher
printf " ---> Setting up flow field \n"
cp $HHOME"/data/build_ic.m" $HHOME"/data/om.dat" $HHOME"/data/eig_"?".dat" .
./matlab_basher.sh build_ic.m " "
rm matlab_basher.sh 
printf " ---> Running simulation ... \n"
./diablo > diablo.out 2>err.out
if [ ! -f end.h5 ]; then echo " Error 4355445. - There have been problems in running diablo. Check the output file of the simulation. "; exit 1; fi
printf " ---> Comparing results ... \n"
cp $HHOME"/data/get_omega.py" .
echo " "$(grep imag om.dat)
python get_omega.py 
printf " <----- CHECK THE AGREEMENT BETWEEN THEORY AND SIMULATIONS!\n"
cd ..
# ---------------------

if [ -d serial ]; then rm -r serial; fi
mkdir serial;cd serial

# USE THE SAME GRID AS THE STABILITY ANALYISIS
cp ../stab_chan/grid.h5 ../stab_chan/grid_def ../stab_chan/diablo .

# ---------------------
# Running serial diablo with HDF5
# ---------------------
printf " Running serial version of diablo ..."
cp $HHOME"/data/input.dat.serial" input.dat 
cp $HHOME"/data/input_chan.dat" input_chan.dat 
./diablo > diablo.out 2>err.out
if [ ! -f end.h5 ]; then echo " Error 4355445. - There have been problems in running diablo. Check the output file of the simulation. "; exit 1; fi
mv out.h5 start.h5
printf "OK\n"
cd ..

module rm $MODULE_SERIAL

# /////////////////////////////
# END SERIAL PART
# /////////////////////////////

# -------------------------
# PARALLEL PART
# -------------------------

module add $MODULE_PARALLEL

if [ -d mpi ]; then rm -r mpi; fi
mkdir mpi;cd mpi

cp $HHOME"/../bin/compDia.sh" . 
cp ../serial/grid_def .
ln -s ../serial/grid.h5 grid.h5
ln -s ../serial/start.h5 start.h5

NPROC=4
# ---------------------
# Generate Parallel HDF5 grid
# ---------------------
printf " Creating parallel HDF5 grid (*.txt format)..."
cp $HHOME"/../bin/genGridParallel.sh" . 
./genGridParallel.sh grid_def $NPROC grid_def_mpi
mv grid_def_mpi grid_def
printf "OK\n"

# ---------------------
# Check for proper compilation MPI
# ---------------------
printf " Compiling parallel version of diablo (HDF5)..."
if [ -f diablo ]; then rm diablo; fi
./compDia.sh -SimDir=$HHOME"/../for" -Clean=yes -Options="HDF5=TRUE PARALLEL=TRUE" > log.out
if [ ! -f diablo ]; then echo " Error 4546876. - Cannot compile parallel version of diablo"; exit 1; fi
printf "OK\n"

# ---------------------
# Running parallel diablo with HDF5
# ---------------------
printf " Running parallel version of diablo ..."
cp $HHOME"/data/input.dat.mpi" input.dat 
cp $HHOME"/data/input_chan.dat" input_chan.dat 
mpirun -np $NPROC ./diablo > diablo.out 2>err.out
if [ ! -f end.h5 ]; then echo " Error 4377745. - There have been problems in running diablo (parallel). Check the output file of the simulation. "; exit 1; fi
printf "OK\n"
cd ..

# ---------------------
# TESTING
# ---------------------
printf " Comparing flow fields between serial and parallel simulations\n"
printf " Computing differences -----> \n"
cp $HHOME"/data/compare_flows.py" .
ln -s serial/grid.h5 grid.h5
python compare_flows.py serial/end.h5 mpi/end.h5
rm grid.h5 compare_flows.py
printf " <----- ERROR SHOULD BE OF THE ORDER OF MACHINE PRECISION!\n "

module rm $MODULE_PARALLEL

printf "\n"
printf " -----------------\n"
printf " END OF THE TESTS.\n"
printf " -----------------\n"
printf "\n"

exit 0 





# Compile the parallel version 
if [ -f diablo ]; then rm diablo; fi
cp ../bin/compDia.sh 
compDia.sh -SimDir=$HHOME"../for" -Clean=yes -Options="HDF5=yes"

if [ ! -f diablo ]; then echo " Error 4343445. - Cannot compile serial version of diablo"; exit(0); fi
